# Proposal for new CommonFilters architecture.
#
# 90% Python, 10% pseudocode in comments; do not try to run this yet!
#
# No imports are defined yet, and this file will be properly split into modules
# in the actual implementation.
#
# There is at least one serious, but not necessarily difficult, problem that must be fixed:
# reconfiguring a FilterStage currently overwrites the configuration of each filter with defaults,
# because it needs to call cleanup(). I'm working on a fix - maybe cleanup() needs to be split
# into "hard" and "soft" variants, with the "soft" variant preserving configuration.
#
# Also importantly, some mechanism is needed so that only the first FilterStage gets the scene,
# while the ones following it get a quad with the output texture of the previous stage.
# I'm trying to figure out a clean solution to this.
#
# Comments welcome.
#
# -JJ


##################################################################
# Filter module
##################################################################

"""This module defines available stages and the Filter API."""

"""Pipeline stages (render passes); ordered list of names considered valid."""
# This is assuming blur represents the camera being out of focus (comes before lens effects)
STAGES = ["main", "blur", "post-blur"]

## Quick reference for Filter interface. See existing filters (below) for complete examples.
#class MyCoolEffectFilter(Filter):
#    def __init__(self, manager):
#        super().__init__(self, manager)
#        # code here
#    def cleanup(self):  # optional
#        # code here
#        super().cleanup(self)
#    def getNeededTextures(self):
#    def getNeedGlow(self):  # optional
#    def getCustomParameters(self):  # optional
#    def getStage(self):  # optional
#    def getSortValue(self):
#    def synthesizeFragmentShaderImpl(self):
#    def createInternalStages(self):  # optional
#    # Then define any filter parameters as properties.

class Filter(object):
    """Abstract base class for postprocessing filters. Filters are implemented in derived classes."""

    def __init__(self, manager):
        """Constructor.

        This calls self.cleanup(), and stores a reference to the manager instance.

        Parameters:
            manager = FilterManager instance being used by the pipeline.
                      Used by filters containing internal stages to set up the
                      internal render-to-quads.

                      This is persistent for the lifetime of the pipeline.

        In your derived classes, be sure to first call super().__init__(self).

        """
        # All methods in Python are virtual, so this will do the right thing.
        self.cleanup()
        # Only the manager is persistent; finalQuad and inputtex are re-created
        # at each reconfigure() of the filter stage.
        self.manager = manager

    def __del__(self):
        """Destructor. Calls self.cleanup() and discards the manager reference."""
        self.cleanup()
        self.manager = None

    def cleanup(self):
        """Cleanup function.

        Resets any dynamically created state.

        By default, this detaches the filter from the pipeline, and sets the needs_compile flag.

        Typically, a derived class should override this to initialize its properties to default values.
        Be sure to call super().cleanup(self) when you're otherwise done.

        See also cleanupInternalStages().

        """
        self.detach()
        self.needs_compile = True

    def attach(self, finalQuad, inputtex):
        """Attach the filter to a pipeline stage.

        This configures the pipeline stage input, and calls createInternalStages().

        This is separate from the constructor, because the required information
        is not available at the creation time of the filter object (filter objects
        are persistent across filter stage reconfigures).

        There is no need to override this in derived classes.

        Parameters:
            finalQuad = NodePath of the output quad from the current stage of the pipeline,
                        into which the compositing fshader renders.

                        This is used to pass custom parameters to the compositing fshader
                        via setShaderInput().

                        Note that it is the finalQuad of *this pipeline stage*.
                        It might not be "the" finalQuad of the whole pipeline,
                        unless this filter resides in the last stage of the pipeline.

            inputtex  = Input texture dictionary, containing input textures for
                        this stage of the pipeline. This is used to access the
                        pipeline stage input in filters containing internal stages.

                        Must contain three entries:
                          "color" -> input color Texture object, or None if not available
                          "depth" -> input depth Texture object, or None if not available
                          "aux"   -> input aux   Texture object, or None if not available

        """
        self.finalQuad = finalQuad
        self.inputtex  = inputtex
        self.createInternalStages()

    def detach(self)
        """Detach the filter from the pipeline.

        This calls cleanupInternalStages(), and sets the pipeline stage input as invalid.

        There is no need to override this in derived classes.

        """
        self.cleanupInternalStages()
        self.finalQuad     = None
        self.inputtex      = {"color" : None, "depth" : None, "aux" : None}

    def getNeededTextures(self):
        """Return a list of textures needed by this filter in the compositing fshader.

        Format: list of tuples (texname, need_texpix)

        Note that since the texture sizes may be different (e.g. internal stages can use
        half-resolution textures if they want to), the need_texpix flag must be specified
        for each texture separately.

        Must be overridden in derived classes.

        Textures available from the pipeline are "color", "depth" and "aux".

        Also any textures defined by the method createInternalStages() can be used here.
        This is useful for filters with internal stages, which need (one or more of)
        their intermediate textures as input to the compositing fshader.

        Including "depth" or "aux" in the list in at least one filter that is enabled
        in the pipeline instructs the pipeline to configure itself to generate them.

        If at least one texture is enabled, l_texcoord will be available in the
        compositing fshader.

        """
        raise NotImplementedError("Filter subclasses must declare which textures they need in the compositing fshader.")

    def getNeedGlow(self):
        """Return whether this filter needs the glow map in the alpha channel of the pipeline stage input color texture.

        Default is False. Can be overridden in derived classes.

        """
        return False

    def getCustomParameters(self):
        """Return a list of custom parameters for the compositing fshader.

        Custom parameters are the k_variables whose values can be changed at runtime using setShaderInput().

        When naming your custom parameters, keep in mind that a large number of different filters
        may get synthesized into the same pipeline stage. The parameter names within each pipeline stage
        must be unique. One way to ensure this is to use a naming convention k_filtername_parametername.

        Format: list of tuples of form (name, type) e.g. ("k_cartooncolor", "float4")

        All custom parameters are "uniform" (in the Cg sense); this is implicit.

        Default is the empty list (no custom parameters). Can be overridden in derived classes.

        """
        return []

    def getStage(self):
        """Return which pipeline stage this filter belongs to.

        Each pipeline stage is a complete render pass.

        The default is 'main'. See Filter.stages. Can be overridden in derived classes.

        """
        return "main"

    def getSortValue(self):
        """Return sort value (priority) within the pipeline stage.

        The sort value is an integer >= 0. Smaller means earlier.

        Must be overridden in derived classes.

        The sort value controls the ordering of the filters within the compositing fshader
        of the pipeline stage.

        """
        raise NotImplementedError("Filter subclasses must declare a sort value.")

    def getNeedCompile(self):
        """Return whether this filter instance currently needs to be (re)compiled.

        There is no need to override this in derived classes.

        In derived classes, the needs_compile flag must be set to True in setters
        for those properties the changing of which requires changes to the synthesized
        fshader code (i.e. compile-time parameters).

        Note that even if needs_compile is False, fragment shader code may be requested
        via synthesizeFragmentShader() if some other filter in the same pipeline stage
        has changed.

        The needs_compile flag is meant to be used such that if, at any given time,
        the flag is False for all filters in a pipeline stage, then at that time
        that stage does not need to be (re)compiled.

        """
        return self.needs_compile

    def synthesizeFragmentShader(self):
        """Synthesize and return fragment shader implementation.

        Interface method.

        There is no need to override this in derived classes.

        This delegates the work to synthesizeFragmentShaderImpl(), and once finished,
        clears the needs_compile flag (which is important).

        The return value is passed through from synthesizeFragmentShaderImpl().

        """
        ret = self.synthesizeFragmentShaderImpl()
        self.needs_compile = False
        return ret

    def synthesizeFragmentShaderImpl(self):
        """Synthesize and return fragment shader implementation.

        Provides the implementation of shader code generation.

        No default. Must be overridden in derived classes.

        The code must respect previous modifications to o_color in the same pipeline stage,
        to avoid overwriting results from other filters earlier in the stage. A filter for which
        this is not possible should be placed as the first one in the pipeline stage (defining
        a new stage, if none of the already defined ones are applicable).

        Return value must be (function_name, code) where both components are strings;
        function_name gives the bare function name which will be called in the compositing fshader
        of the pipeline stage, and code is the complete source code of the Cg function as a string,
        however *without* the function declaration and surrounding braces.

        Both the function declaration and the call to the function in the compositing fshader
        are generated automatically by the pipeline, using the data obtained from getNeededTextures()
        and getCustomParameters().

        The pipeline automatically composes the final compositing fshader from the
        code defined by the enabled filters. The function calls are ordered using
        the sort value of the filters.

        The implementation must use in the synthesis the current values of any compile-time parameters
        stored in the properties of the derived class.
        
        """
        raise NotImplementedError("Filter subclasses must implement fragment shader synthesis.")

    def createInternalStages(self):
        """Set up internal processing stages, if needed.

        For example, a blur filter might have internal "blur-x" and "blur-y" passes
        that are applied before the compositing fshader.

        This method is optional; the default implementation is blank.

        """
        pass

    def cleanupInternalStages(self):
        """Clean up internal processing stages, if needed.

        By default, this sets self.textures and self.interQuads to blank.
        Usually this is sufficient.

        If you additionally need something else, override this, but be sure to
        call super().cleanupInternalStages(self) when you're otherwise done.

        """
        self.textures = {}
        self.interQuads = []


##################################################################
# Example filters
##################################################################

# The more complex filters will be in their own modules; short, simple
# filters will be collected into a MiscFilters module.

class ColorInitializationFilter(Filter):
    """This filter simply initializes o_color from the current pixel in the stage input color texture.

    It is inserted by default as the first filter in each pipeline stage.

    """
    def __init__(self, manager):
        super().__init__(self, manager)

    def getNeededTextures(self):
        # compositing shader needs color texture, but no texpix.
        return [ ("color", False) ]

    def getStage(self):
        # This filter is special; a copy of it goes into all stages.
        return None

    def getSortValue(self):
        return -1  # initialization goes first before any other filter.

    def synthesizeFragmentShaderImpl(self):
        return ("initializeOutputColor", "o_color = tex2D(k_txcolor, l_texcoord.xy);\n")


# A simple filter. Examples of:
#  - Storing filter parameters in properties
#  - Compile-time parameter
#
class TintFilter(Filter):
    """Color tint / fade-to-color filter.

    In tint mode [default]:

      The color of each pixel is tinted (multiplied) by the specified constant color.
      The result is saturated to [0,1] in each component.

      The tinted pixel then is blended with the original one with the strength
      specified by blend.

    In fade mode:

      Each pixel is blended toward the specified constant color.
      The parameter blend controls the strength.

    This filter can be used e.g. to fade the screen to black (color=(0,0,0,1))
    by dynamically updating the blend parameter in your own update task
    so that it gradually goes from 0 to 1.

    The fade mode also allows fading toward colors other than black
    (e.g. color=(1,1,1,1) to fade to white).

    In tint mode, fading the color to (255,255,255,1) gives another kind of
    fade to white (using the fact that there are 8 bits per color channel).

    Parameters:

      mode  = string, one of:
                "tint" = tint (multiply) by the specified color.
                "fade" = fade to the specified color.

      color = (R,G,B,A) tuple; tint color. To make a component no-op, set it to 1.
              Usually setting the alpha component to 1 is a good idea.

      blend = strength of blending between the original and processed color.
              1.0 means fully tinted or faded, 0.0 does nothing.

    """
    def __init__(self, manager):
        super().__init__(self, manager)  # this will call self.cleanup()

    def cleanup(self):
        self._color = (0.5,0.5,0.5,1)
        self._blend = 1.0
        self._mode  = "tint"  # compile-time parameter
        super().cleanup(self)

    def getNeededTextures(self):
        return [ ("color", False) ]

    def getCustomParameters(self):
        return [ ("k_tint_tint_color", "float4"),
                 ("k_tint_blend",      "float") ]

    def getSortValue(self):
        return 90  # late within the pipeline stage

    def synthesizeFragmentShaderImpl(self):
        code = "// Color tint / fade-to-color filter\n"
        if self.mode == "tint":
            code += "//   Mode: color tint\n"
            code += "float4 tinted_o_color = saturate(o_color * k_tint_tint_color);\n"
            code += "o_color = lerp(o_color, tinted_o_color, k_tint_blend);\n"
        elif self.mode == "fade":
            code += "//   Mode: fade-to-color\n"
            code += "o_color = lerp(o_color, k_tint_tint_color, k_tint_blend);\n"
        else:
            raise ValueError("TintFilter: invalid mode '%s'; valid: 'tint', 'fade'" % self.mode)
        return ("tintFilter", code)

    @property
    def color(self):
        """Tint color as (R,G,B,A) float tuple, each component in [0,1]."""
        return self._color
    @color.setter
    def color(self, value):
        self._color = value
        self.finalQuad.setShaderInput("tint_tint_color", self._color)

    @property
    def blend(self):
        """Effect strength, float. 1 is full strength, 0 is off."""
        return self._blend
    @blend.setter
    def blend(self, value):
        self._blend = value
        self.finalQuad.setShaderInput("tint_blend", self._blend)

    @property
    def mode(self):
        """Effect mode, either 'tint' or 'fade'."""
        return self._mode
    @mode.setter
    def mode(self, value):
        self._mode = value
        self.needs_compile = True


# A filter that needs texpix.
#
class ScanlinesFilter(Filter):
    FIELD_TOP    = 0
    FIELD_BOTTOM = 1

    def __init__(self, manager):
        super().__init__(self, manager)

    def cleanup(self):
        self._field = self.FIELD_TOP
        self._blend = 0.5
        self._line_thickness = 1
        super().cleanup(self)

    def getNeededTextures(self):
        return [ ("color", True) ]  # needs color texture, and texpix for it

    def getCustomParameters(self):
        return [ ("k_scanlines_field",     "int"),
                 ("k_scanlines_blend",     "float"),
                 ("k_scanlines_thickness", "int") ]

    def getStage(self):
        return "post-blur"

    def getSortValue(self):
        return 100  # should probably be applied last

    def synthesizeFragmentShaderImpl(self):
        code  = "// CRT TV scanlines filter\n"
        code += "if (int(l_texcoordC.y / texpix_txcolor.y)/k_scanlines_thickness % 2 != k_scanlines_field)\n"
        code += "  o_color.rgb = (1.0 - k_scanlines_blend) * o_color.rgb;\n"

    @property
    def field(self):
        """Which field to keep, either ScanlinesFilter.FIELD_TOP or ScanlinesFilter.FIELD_BOTTOM."""
        return self._field
    @field.setter
    def field(self, value):
        self._field = value
        self.finalQuad.setShaderInput("scanlines_field", self._field)

    @property
    def blend(self):
        """Effect strength (float). 1 is fully scanlined, 0 does nothing."""
        return self._blend
    @blend.setter
    def blend(self, value):
        self._blend = value
        self.finalQuad.setShaderInput("scanlines_blend", self._blend)

    @property
    def line_thickness(self):
        """Thickness of each scanline in pixels (integer)."""
        return self._line_thickness
    @line_thickness.setter
    def line_thickness(self, value):
        self._line_thickness = value
        self.finalQuad.setShaderInput("scanlines_thickness", self._line_thickness)


# A filter with internal stages.
#
class BlurFilter(Filter):
    """A blur/sharpen filter. If the 'amount' parameter is 1.0, it will not have effect.
    A value of 0.0 means fully blurred, and a value higher than 1.0 sharpens the image."""

    def __init__(self, manager, finalQuad, inputtex):
        super().__init__(self, manager, finalQuad, inputtex)

    def cleanup(self):
        self._blurval = 0.0  # full strength
        super().cleanup(self)

    def getNeededTextures(self):
        # The compositing shader blends between the color texture and the internal blur1 texture.
        return [ ("color", False), ("blur1", False) ]

    def getCustomParameters(self):
        return [ ("k_blurval", "float") ]

    def getStage(self):
        # Input texture to the pipeline stage where the blur filter is placed
        # must see all previous changes; hence, this must come after main stage.
        return "blur"

    def getSortValue(self):
        return 0  # first within the pipeline stage (right after init)

    def synthesizeFragmentShaderImpl(self):
        code  = "// Blur/sharpen blend pass\n"
        code += "o_color = lerp(tex2D(k_txblur1, l_texcoord.xy), o_color, k_blurval.x);\n"
        return ("blurFilter", code)

    def createInternalStages(self):
        # create textures for internal stages
        #
        for tex in ["blur0", "blur1"]
            self.textures[tex] = Texture("scene-"+tex)
            self.textures[tex].setWrapU(Texture.WMClamp)
            self.textures[tex].setWrapV(Texture.WMClamp)

        # set up shaders for internal stages
        #
        blur0 = self.textures["blur0"]
        blur1 = self.textures["blur1"]
        self.interQuads.append(self.manager.renderQuadInto(colortex=blur0,div=2))
        self.interQuads.append(self.manager.renderQuadInto(colortex=blur1))
        self.interQuads[0].setShaderInput("src", self.inputtex["color"])
        self.interQuads[0].setShader(self.loadShader("filter-blurx.sha"))
        self.interQuads[1].setShaderInput("src", blur0)
        self.interQuads[1].setShader(self.loadShader("filter-blury.sha"))

    @property
    def blurval(self):
        """Effect strength, float. 1 is off; values in [0,1) blur, values in (1,2] sharpen."""
        return self._blurval
    @blurval.setter
    def blurval(self, value):
        self._blurval = value
        self.finalQuad.setShaderInput("blurval", self._blurval)


# A filter that needs the glow map in the alpha channel of the input color texture.
#
class ViewGlowFilter(Filter):
    def __init__(self, manager, finalQuad, inputtex):
        super().__init__(self, manager, finalQuad, inputtex)

    def getNeededTextures(self):
        return []

    def getNeedGlow(self):
        return True

    def getStage(self):
        return "post-blur"  # this is a debug helper filter which should go last

    def getSortValue(self):
        return 100  # last thing in the stage

    def synthesizeFragmentShaderImpl(self):
        code  = "// ViewGlow (debug filter)\n"
        code += "o_color.r = o_color.a;\n"
        return ("viewGlowFilter", code)


##################################################################
# Pipeline logic
##################################################################

# A FilterPipeline is made of FilterStages.

class FilterStage(object):
    def __init__(self, manager):
        self.manager = manager  # FilterManager instance
        self.filters = []  # Filter subclass instances

    def cleanup(self):
        for f in self.filters:
            f.cleanup()  # this will detach() the filters, too
        self.filters = []  # Filter subclass instances
        self.manager.cleanup()
        self.finalQuad = None
        self.stageInputTextures = {}

    def findFilter(self, fclass):
        """Find filter in this FilterStage by its object type (class).

        Example:
            i = myFilterStage.findFilter(ScanlinesFilter)

        Return value:
            Index of filter in self.filters if found, None if not.

        """
        found = False
        for i,f in enumerate(self.filters):
            if isinstance(f, fclass):
                found = True
                break
        if found:
            return i
        else:
            return None

    def setFilter(self, f):
        """Set (enable) a filter.

        The parameter f must be an instance of a Filter subclass.

        Typically at this point the filter object has been configured (by setting its properties),
        but it is not attached to a FilterStage.

        If the filter of the given type is not present in the FilterStage, it is added.
        This is the recommended usage.


        If an instance of the same type of filter is already present, its configuration is updated.
        This is done by querying all properties, and copying changed ones from f to the existing instance.

        Copying only the changed properties ensures that this won't trigger a shader recompile
        unless it has to.

        This is a hack to support the old CommonFilters interface, which used to replace the
        entire configuration of a filter at one go.

        """
        if not isinstance(f, Filter):
            raise ValueError("f must be an instance of a Filter subclass, but is '%s'" % type(f))

        oldidx = self.findFilter(type(f))
        if oldidx is None:
            self.filters.append(f)
            self.reconfigure()  # this will sort the filters
        else:
            # Copy configuration only.
            #
            # We assume here that all configuration parameters are wrapped into properties.
            #
            # Scan all properties of the class f belongs to. (This should also find properties
            # defined in any of f's ancestors.)
            #
            import inspect  # TODO: move this to the start of the module
            def isproperty(v):
                return isinstance(v, property)
            propertynames = [name for (name, value) in inspect.getmembers(type(f), isproperty)]

            # We only want to trigger the setter if the value of the property has changed.
            #
            # This is because setting a property might set the needs_compile flag;
            # by design, FilterStage doesn't know which parameters are compile-time.
            #
            fcurr = self.filters[oldidx]
            for s in propertynames:
                newvalue = getattr(f,s)
                if newvalue != getattr(fcurr,s):
                    setattr(fcurr,s,newvalue)

            # Reset this FilterStage if the filter needs a recompile.
            #
            if fcurr.getNeedCompile():
                self.reconfigure()

    def delFilter(self, fclass):
        """Remove a filter from this FilterStage.

        There can be only one of each type of filter in the stage,
        so the removal is requested by object type (class).

        If fclass specifies a filter type that is not present in this FilterStage,
        then this does nothing.

        """
        i = self.hasFilter(type(f))
        if i is None:
            self.filters[i].cleanup()
            del self.filters[i]
            # It may happen none of the remaining filters wants a recompile at the moment.
            # Thus, we force the rebuild to ensure the deleted filter gets removed.
            self.reconfigure(force=True)

    def sortFilters(self):
        """Sort filters in this pipeline stage by their sort value.

        Used by reconfigure().

        """
        def cmpBySortValue(f1,f2):
            i1 = f1.getSortValue()
            i2 = f2.getSortValue()
            if i1 < i2:
                return -1
            elif i1 > i2:
                return +1
            else:
                return 0
        self.filters.sort(cmp=cmpBySortValue)

    def reconfigure(self, force=False):
        """Rebuild shaders.

        Usually this works intelligently, only recompiling shaders when at least one filter
        reports that it needs to be recompiled.

        This behavior can be overridden by setting force=True.

        """
        # This function always works like the "fullrebuild" mode in old CommonFilters (up to 1.8.1).
        #
        # Because configuration now resides in filter object properties, and they take care of
        # updating the appropriate shader inputs, FilterStage does not need to do anything
        # when the configuration changes, unless a shader recompile is needed.

        # Check that we need a compile before doing anything.
        #
        if not force:
            needsCompile = False
            for f in self.filters:
                if f.getNeedCompile():
                    needsCompile = True
                    break
            if not needsCompile:
                return True  # no compile needed, we're done

        # Clean up, ensure correct filter ordering, and re-initialize.
        #
        self.cleanup()  # this will also detach() our filters by calling their cleanup
        self.sortFilters()

        # Create input textures for this pipeline stage and set up auxbits
        #
        # Any one of these that is not needed by any filter remains None.
        self.stageInputTextures = { "color" : None,
                                    "depth" : None,
                                    "aux"   : None }
        auxbits = 0
        needscenetex = {}
        for f in self.filters:
            f_needtex = map( lambda item: item[0], f.getNeededTextures() )  # names only
            for tex in f_needtex:
                needscenetex[tex] = True
            if f.getNeedGlow():
                auxbits |= AuxBitplaneAttrib.ABOGlow
        if "aux" in all_needtex:
            auxbits |= AuxBitplaneAttrib.ABONormal
        for tex in needscenetex:
            self.stageInputTextures[tex] = Texture("scene-"+tex)
            self.stageInputTextures[tex].setWrapU(Texture.WMClamp)
            self.stageInputTextures[tex].setWrapV(Texture.WMClamp)

        # Create finalQuad
        #
        # TODO: we might not have a scene here (if this is not the first FilterStage)
        #
        # TODO: also, arguably this level of initialization belongs in FilterPipeline,
        #       which knows which FilterStage is which and how they should be connected.
        #       But it requires the auxbits and (blank) texture objects, which are the
        #       most natural to set up here. Think about this part some more.
        #
        self.finalQuad = self.manager.renderSceneInto(textures=self.textures, auxbits=auxbits)
        if (self.finalQuad == None):
            self.cleanup()
            return False

        # Attach filters to the stage.
        #
        # We must redo this each time reconfigure() is called, because finalQuad and
        # stageInputTextures are re-created.
        #
        for f in self.filters:
            f.attach(self.finalQuad, self.stageInputTextures)

        # Synthesize compositing shader for this pipeline stage
        cgHeader = "//Cg\n"

        # compose vshader
        vshader = "void vshader(float4 vtx_position : POSITION,\n"
        vshader += " out float4 l_position : POSITION,\n"
        # TODO: fill in parameters based on getNeededTextures()
        vshader += " uniform float4x4 mat_modelproj)\n"
        vshader += "{\n"
        vshader += " l_position=mul(mat_modelproj, vtx_position);\n"
        # TODO: fill in texcoords using rdb's optimized solution
        # TODO: HalfPixelShift (FilterStage property)
        vshader += "}\n"

        # compose filter functions
        # TODO: getNeededTextures(), getCustomParameters(), synthesizeFragmentShader()

        # compose fshader
        fshader = "void fshader(\n"
        for f in self.filters:
            # TODO: build parameter list: getNeededTextures() (also texpix!), getCustomParameters()
        fshader += "out float4 o_color : COLOR)\n"
        fshader += "{\n"
        for f in self.filters:
            # TODO: insert filter function calls
            #       getNeededTextures(), getCustomParameters(), synthesizeFragmentShader()
        fshader += "}\n"

        # Compose complete Cg program, and compile the resulting compositing shader.
        #
        program = cgHeader + vshader + filterFunctions + fshader
        self.finalQuad.setShader(Shader.make(program))

        # Set texture inputs to the compositing shader.

        # scene textures (provide only those scene textures that are needed by at least one filter)
        for tex in needscenetex:
            self.finalQuad.setShaderInput("tx"+tex, self.stageInputTextures[tex])

        # internal textures defined by the filters
        for f in self.filters:
            # check that tex is in the list of textures the filter wants in the compositing shader
            needtex = map( lambda item: item[0], f.getNeededTextures() )  # names only
            for tex in f.textures.keys():  # f.textures will be blank for most filters (no internal stages)
                if tex in needtex:
                    self.finalQuad.setShaderInput("tx"+tex, f.textures[tex])

        # TODO: where does this belong in the new architecture?        
        #self.task = taskMgr.add(self.update, "common-filters-update")
        #self.update()

        return True

    # TODO: getter for output textures (color, depth, aux)
    #       depth and aux can be passed through for now, but next stage in pipeline needs modified color


class FilterPipeline(object):
    # TODO
    # - create a FilterManager instance (at __init__(self, win, cam))
    # - reconfigure(): examine configuration (iterate over filter instances in self.filters),
    #   create the appropriate FilterStages, assign our Filters to them
    # - pass the FilterManager instance to each FilterStage
    # - we here add filters to the correct FilterStage; FilterStage takes care of attach()
    # - during a pipeline reconfigure, all filters are detach()ed


class CommonFilters:  # note that the original was not a new-style class, hence this isn't either.
    # TODO
    # backward-compatible API goes here
    #  - manage one instance of FilterPipeline

